// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


/**
   Realiza peticiones automaticas sobre el servidor HTTP de operaciones aritmeticas.

   La cadencia de envio de mensajes se establece mediante un temporizador.
   Los operadores se calculan de forma aleatoria.

   El servidor de este cliente: http_server.p o http_rserver.p
*/

#include <iostream>

#include <string.h>
#include <unistd.h>

#include <anna/core/core.hpp>
#include <anna/app/functions.hpp>
#include <anna/comm/comm.hpp>

#include <anna/xml/DocumentMemory.hpp>
#include <anna/xml/Node.hpp>
#include <anna/xml/Parser.hpp>

#include <anna/timex/Engine.hpp>
#include <anna/timex/Clock.hpp>

#include <anna/http/Request.hpp>
#include <anna/http/Response.hpp>
#include <anna/http/Transport.hpp>
#include <anna/http/Handler.hpp>

static const Millisecond Resolution(250);
static const Millisecond Period(500);
static const Millisecond OneSecond(1000);

class Sender : public anna::timex::Clock {
public:
   Sender ();
   
   void setMessageBySecond (const int messageBySecond) throw () { a_messageByTick = messageBySecond / (OneSecond / Period); }

   int getTxMessageCounter () const throw () { return a_txMessageCounter; }

private:
   struct Attribute { enum _v { ValueOne, ValueTwo, Operator, Time, Max }; };
   
   int a_messageByTick;
   int a_nquarter;
   int a_errorCounter;
   int a_txMessageCounter;
   http::Request a_httpRequest;
   xml::Node* a_xmlRequest;
   xml::Attribute* a_xmlAttributes [Attribute::Max];

   /* Se invoca 4 veces por segundo */
   bool tick () throw (RuntimeException);
};

class MyHandler : public http::Handler {
public:
   MyHandler () : http::Handler ("http_client::MyHandler") {;}

private:
   xml::DocumentMemory a_xmlRequest;
   xml::Parser a_xmlParser;
   
   void evRequest (ClientSocket&, const http::Request&) throw (RuntimeException) {;}
   void evResponse (ClientSocket&, const http::Response&) throw (RuntimeException);
   
//   static bool isOk (const test::Response& response) throw ();
};

class MyCommunicator : public Communicator {
public:
   MyCommunicator () : Communicator (), a_avgResponseTime (0), a_rxMessageCounter (0) {;}

   void count (const int delay) throw (RuntimeException);

private:
   int a_avgResponseTime;
   int a_rxMessageCounter;
   MyHandler a_httpHandler;

   
   void eventReceiveMessage (ClientSocket&, const Message&) throw (RuntimeException);

   void eventBreakConnection (const ClientSocket&) throw ();

   // Sustituye la redefinicin de los siguientes mtodos
   using comm::Communicator::eventBreakConnection;
/*
   void eventBreakConnection (Server* server) throw () {
      comm::Communicator::eventBreakConnection (server);
   }
   void eventBreakConnection (const Service* service) throw () { 
      comm::Communicator::eventBreakConnection (service);
   }
*/      
};

class HeavyWIMS20Client : public anna::comm::Application {
public:
   HeavyWIMS20Client ();

   Server* getServer () const throw () { return a_server; }
   const Sender* getSender () const throw () { return &a_sender; }

private:
   MyCommunicator a_communicator;
   anna::timex::Engine a_timeController;
   Sender a_sender;
   Server* a_server;

   void initialize () throw (RuntimeException);
   void run () throw (RuntimeException);
};

using namespace std;

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   HeavyWIMS20Client app;

   srand (time (NULL));

   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      Logger::setLevel (Logger::Information);
      string traceFile ("client.");
      traceFile += anna::functions::asString ((int) getpid ());
      traceFile += ".trace";
      Logger::initialize ("http_client", new TraceWriter (traceFile.c_str (),4096000));

      app.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }

   return 0;
}

HeavyWIMS20Client::HeavyWIMS20Client () :
   Application ("http_client", "Cliente HTTP", "1.0"),
   a_communicator (),
   a_timeController (OneSecond, Resolution)
{
   CommandLine& commandLine (CommandLine::instantiate ());

   commandLine.add ("p", CommandLine::Argument::Mandatory, "Puerto en el que el servidor atiende respuestas.");
   commandLine.add ("a", CommandLine::Argument::Mandatory, "Direccin IP Puerto en el que el servidor atiende respuestas.");
   commandLine.add ("n", CommandLine::Argument::Mandatory, "Numero de mensajes por segundo");
   commandLine.add ("trace", CommandLine::Argument::Optional, "Nivel de trazas (debug,warning, error,...)");
}

void HeavyWIMS20Client::initialize ()
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   Network& network = Network::instantiate ();

   a_server = network.createServer (cl.getValue ("a"), cl.getIntegerValue ("p"), true, &http::Transport::getFactory ());
   a_sender.setMessageBySecond (cl.getIntegerValue ("n"));

   if (cl.exists ("trace"))
      Logger::setLevel (Logger::asLevel (cl.getValue ("trace")));
}

void HeavyWIMS20Client::run ()
   throw (RuntimeException)
{
   a_timeController.activate (a_sender);

   a_communicator.accept ();
}

void MyCommunicator::eventReceiveMessage (ClientSocket& clientSocket, const Message& message)
   throw (RuntimeException)
{
   LOGMETHOD (TraceMethod tm ("MyCommunicator", "eventReceiveMessage", ANNA_FILE_LOCATION));

   if (clientSocket.support (http::Transport::className ()) == false)
      return;

   a_httpHandler.apply (clientSocket, message);
}

void MyCommunicator::count (const int delay)
   throw (RuntimeException)
{
   Guard guard (this, "MyCommunicator::count");
   
   a_rxMessageCounter ++;
   a_avgResponseTime += delay;
}

void MyCommunicator::eventBreakConnection (const ClientSocket& clientSocket)
   throw ()
{
   if (a_rxMessageCounter == 0)
      return;

   LOGNOTICE (
      HeavyWIMS20Client& app = static_cast <HeavyWIMS20Client&> (anna::app::functions::getApp ());   
      string msg ("Tiempo medio respuesta: ");
      msg += anna::functions::asString (a_avgResponseTime / a_rxMessageCounter);
      msg += " ms";
      msg += anna::functions::asText (" | Rx: ", a_rxMessageCounter);
      msg += anna::functions::asText (" | Tx: ", app.getSender ()->getTxMessageCounter ());
      Logger::notice (msg, ANNA_FILE_LOCATION);

      cout << msg << endl << endl;
   );
   requestStop ();
   comm::Communicator::eventBreakConnection (clientSocket);
}

Sender::Sender () : Clock ("Sender", Period), 
   a_messageByTick (0), 
   a_nquarter (0), 
   a_errorCounter (0),
   a_txMessageCounter (0)
{
   a_httpRequest.setMethod (http::Method::Type::Get);
   a_httpRequest.setURI ("http_xmlclient.p");
   
   /**
    * Crea el documento XML que usaremos para codificar la respuesta y 
    * pre-localiza los objetos sobre los que tendr que actuar
    */
   a_xmlRequest = new xml::Node ("Request");
   a_xmlAttributes [Attribute::ValueOne] = a_xmlRequest->createAttribute ("ValueOne", 0);
   a_xmlAttributes [Attribute::ValueTwo] = a_xmlRequest->createAttribute ("ValueTwo", 0);
   a_xmlAttributes [Attribute::Operator] = a_xmlRequest->createAttribute ("Operator", 0);
   a_xmlAttributes [Attribute::Time] = a_xmlRequest->createAttribute ("Millisecond", 0);
}

bool Sender::tick ()
   throw (RuntimeException)
{
   Server* server = static_cast <HeavyWIMS20Client&> (anna::app::functions::getApp ()).getServer ();
   Communicator* communicator = anna::app::functions::component <Communicator> (ANNA_FILE_LOCATION);

   if (a_errorCounter > 100) {
      communicator->requestStop ();
      Logger::warning ("Terminado por errores continuos en la conexion", ANNA_FILE_LOCATION);
      return false;
   }

   for (int n = 0; n < a_messageByTick && communicator->hasRequestedStop () == false; n ++) {
      a_xmlAttributes [Attribute::ValueOne]->setValue (rand () % 1000);
      a_xmlAttributes [Attribute::ValueTwo]->setValue (rand () % 1000);
      a_xmlAttributes [Attribute::Operator]->setValue ("+");
      a_xmlAttributes [Attribute::Time]->setValue (anna::functions::millisecond ());

      try {
         a_httpRequest.setBody (a_xmlRequest);
         server->send (a_httpRequest);
         a_txMessageCounter ++;
      }
      catch (RuntimeException& ex) {
         a_errorCounter ++;
         ex.trace ();
         break;
      }
   }

   return true;
}

void MyHandler::evResponse (ClientSocket& clientSocket, const http::Response& response)
   throw (RuntimeException)
{
   if (response.getStatusCode () != 200) 
      return;
      
   a_xmlRequest.initialize (response.getBody ());

   const xml::Node* root = a_xmlRequest.parse ();
   
   const anna::Millisecond now = anna::functions::millisecond ();
   const anna::Millisecond past(root->getAttribute ("Time")->getIntegerValue ()); 
   
   const int delay = now - past; 

   if (delay < 0) {
      LOGWARNING (
         string msg = a_xmlRequest.getContentAsCString ();
         msg += anna::functions::asText (" | Delay: ", delay);
         Logger::warning (msg, ANNA_FILE_LOCATION);
      );
      return;
   }

   app::functions::component <MyCommunicator> (ANNA_FILE_LOCATION)->count (delay);
}
