// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_LargeBinaryCodec_hpp
#define anna_comm_LargeBinaryCodec_hpp

#include <vector>

#include <anna/core/DataBlock.hpp>

#include <anna/comm/Codec.hpp>

namespace anna {

namespace comm {

/**
   Codificador/Decodificador de objetos complejos.

   A diferencia de EnhancedCodec que esta pensado para seriealizar objetos con estructuras
   tipos de datos basicos, este codec esta pensado especialmente para serializar objetos o
   listas de objetos tan complejas como sea preciso.

   El mensaje estara formado por una serie indeterminada de bloques de datos binarios, lo que
   nos permite serializar facilmente cualquier tipo de estructura de datos.

   \warning Esta clase no establece proteccion ante accesos concurrentes
*/
class LargeBinaryCodec : public Codec {
public:
  typedef std::vector <DataBlock*>::iterator iterator;
  typedef std::vector <DataBlock*>::const_iterator const_iterator;

  /**
     Constructor.

     @param type Tipo por el que sera conocido este tipo de mensaje.
     @param scramble Indica si el mensaje debe ser codificado de forma que no se pueda ver el contenido
     del mismo con una simple herramienta de monitorizacin de mensajes de red. Por defecto esta
     activo ya que la codificacin realizada es muy simple y rapida y el tiempo empleado es casi
     inapreciable.
  */
  LargeBinaryCodec(const Type type, const bool scramble = true) :
    Codec(type, scramble),
    a_dataBlock(true) {
    attach("LargeBinaryBlock", a_dataBlock);
  }

  /**
     Destructor.
  */
  virtual ~LargeBinaryCodec() { reset(); clear(); }

  /**
     Elimina el contenido del mensaje asociado a este objeto.
  */
  void clear() throw(RuntimeException) { a_dataBlock.clear(); }

  /**
     Amplia el contenido del mensaje con el contenido del bloque de memoria recibido.
     \param dataBlock Bloque de memoria con el que vamos a ampliar el mensaje.
     \return La referencia a este objeto.
  */
  LargeBinaryCodec& operator += (const DataBlock& dataBlock) throw(RuntimeException);

  /**
     Transfiene la informacin contenida en el bloque de memoria recibido hacia las variables asociadas a este mensaje interno.
     \param dataBlock Bloque de memoria que contiene las variables codificadas.
     Para acceder la lista de bloques que componen el mensaje recibido se deben utilizar los metodo begin y end.
  */
  virtual void decode(const DataBlock& dataBlock) throw(RuntimeException);

  /**
     Devuelve el iterador que apunta al comienzo de la serializacin.
     \return El iterador que apunta al comienzo de la serializacin.
     \warning El valor de este metodo solo sera valido despues de invocar  a decode.
  */
  const_iterator begin() const throw() { return a_blocks.begin(); }

  /**
     Devuelve el iterador que apunta al final de la serializacin.
     \return El iterador que apunta al final de la serializacin.
     \warning El valor de este metodo solo sera valido despues de invocar  a decode.
  */
  const_iterator end() const throw() { return a_blocks.end(); }

  /**
   * Devuelve el nmero de bloques contenidos en el mensaje.
   * \return el nmero de bloques contenidos en el mensaje.
   */
  int size() const throw() { return a_blocks.size(); }

  /**
     Devuelve el puntero al objeto referenciado por el iterador recibido como parametro.
     \param ii Iterador.
     \return el puntero al objeto referenciado por el iterador recibido como parametro.
  */
  static const DataBlock* data(const_iterator ii) throw() { return *ii; }

  /**
     Devuelve el puntero al objeto referenciado por el iterador recibido como parametro.
     \param ii Iterador.
     \return el puntero al objeto referenciado por el iterador recibido como parametro.
  */
  static DataBlock* data(iterator ii) throw() { return *ii; }

private:
  DataBlock a_dataBlock;
  std::vector <DataBlock*> a_blocks;

  void reset() throw();
};

}
}

#endif

