// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_tracing_Configuration_hpp
#define anna_core_tracing_Configuration_hpp



#include <anna/core/Singleton.hpp>
#include <anna/core/tracing/Logger.hpp>

#include <anna/core/util/RegularExpression.hpp>

// STL
#include <string>
#include <map>



namespace anna {

namespace tracing {


/**
* Class used for selective tracing configuration
*/
class Configuration : public anna::Singleton <Configuration> {


public:

  typedef std::map<anna::RegularExpression, anna::Logger::Level> trace_trigger_map_t;

private:

  trace_trigger_map_t a_traceTriggers; // configured triggers map, for selective tracing
  bool a_readTraceTriggersAsRegexp; // boolean to configure the way in which trace triggers are interpreted


  Configuration() { a_readTraceTriggersAsRegexp = false; }  // private constructor

  friend class anna::Singleton <Configuration>;
  friend class TraceLevelChecker;


  const trace_trigger_map_t & getTraceTriggers() const throw() { return a_traceTriggers; }


public:

  /**
  * Deletes all added triggers for selective tracing.
  */
  void resetTraceTriggers() throw() { a_traceTriggers.clear(); }


  /**
  * Activates a trigger for selective tracing. The normal way is to activate only one
  * (usually applications register MSISDN, but we prefer the name 'trigger' because
  * selective tracing capabilities are generalized for any context data we want to detect).
  * Anyway, it is possible to add several string references for multiple context log activation.
  *
  * @param trigger String reference for contexts comparison
  * @param accumulate Boolean for trigger filter accumulation over internal set. I.e., application
  * could test any MSISDN within a configured set. False by default, means activation for only one
  * trigger reference.
  * @param level Desired trace level when context fulfill this trigger reference. Debug by default
  */
  void activateTraceTrigger(const std::string & trigger, bool accumulate = false, anna::Logger::Level level = anna::Logger::Debug) throw();


  /**
  * Deletes a trigger for selective tracing.
  *
  * @param trigger String reference to be erased from internal trigger set
  */
  void deactivateTraceTrigger(const std::string & trigger) throw();


  /**
  * Interpret all trace triggers as regular expressions.
  * Is used from TraceLevelChecker to provide strict or regexp-based checkings.
  * If not configured, regexp is disabled.
  *
  * @return Boolean about if regular expressions interpretation are activated.
  */
  bool readRegexpForTraceTriggers() const throw() { return (a_readTraceTriggersAsRegexp); }


  /**
  * Configure all trace triggers to be interpreted as regular expressions.
  * Application can set this depending on trace filtering needs.
  * If not configured, regexp is disabled for TraceLevelChecker.
  *
  * @param readRegexp Boolean for activation/deactivation of regexp funtionality
  */
  void useRegexpForTraceTriggers(bool readRegexp = true) throw() { a_readTraceTriggersAsRegexp = readRegexp; }


  /**
  * Class string representation
  *
  * @return String with class content
  */
  std::string asString(void) const throw();
};

}
}

#endif

