// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_codec_basetypes_AvpData_hpp
#define anna_diameter_codec_basetypes_AvpData_hpp


#include <anna/core/RuntimeException.hpp>
#include <anna/core/DataBlock.hpp>
#include <anna/core/functions.hpp>

#include <anna/core/functions.hpp>

// STL
#include <string>


namespace anna {

namespace diameter {

namespace codec {

namespace basetypes {

/**
* Diameter avp abstract data container
*
* All diameter types inherit from this class. Basic types with one level, derived ones with two levels.
* Three or more inheritance levels have no sense. Each level has own private members and updateBasic()
* will be implemented at derived types to keep coherence towards parent ones. This coherence implies to
* call updateBasic when private members are updated.
*
*<pre>
* Basic diameter types are:      OctetString, Integer32, Integer64, Unsigned32, Unsigned64, Float32, Float64
* Derived diameter types are:    Address, Time, UTF8String, DiameterIdentity, DiameterURI, IPFilterRule, QoSFilterRule (derived from OctetString)
*                                Enumerated (derived from Integer32)
* Printable derived types are:   UTF8String, DiameterIdentity, DiameterURI, IPFilterRule, QoSFilterRule
*
* User application could create new diameter types derived for any basic one, and reimplement the methods to adjust its behaviour.
*
*</pre>
*
* There are setters/getters at any level, and a lot of helpers to decode/interpret the stored data.
* Then, this class could be used as container and/or helper tool.
* Parent setters won't modify child members and vice versa, but helpers syncronizes both levels through parent 'code()' and child 'decode()' methods.
*
* Derived types with own members, must use protected or private inheritance, and export base class methods desired (using declaration).
* This mechanism guarantees private members coherence between different levels: a derived type instance couldn't directly manipulate base class own members.
*/
class AvpData {

  /**
  * Encodes the avp data part over target buffer.
  * THIS WILL BE IMPLEMENTED AT BASIC TYPES because works with parent members
  *
  * @param buffer Raw data to be encoded
  * @param size Size of raw data to be encoded
  */
  virtual void codeBasic(char* buffer, int &size) throw(anna::RuntimeException) = 0;

  /**
  * Updates parent members from child ones to keep coherence between levels.
  * Neither of diameter basic types need to be checked and have no parents,
  * then THIS WILL BE IMPLEMENTED AT DERIVED TYPES WITH NEW MEMBERS (not, i.e. at Enumerated)
  * and will be called after child members modification (setters).
  * Basic types implementation will be empty.
  */
  virtual void updateBasic() throw(anna::RuntimeException) {;}

  /**
  * Sets own members from natural/smart string representation
  * String argument is never provided NULL (internally checked)
  *
  * @param printableString avp data in natural/smart string representation (human-readable)
  */
  virtual void setPrintableString(const char * printableString) throw(anna::RuntimeException) = 0;


protected:

  /**
  * Asserts printable nature for buffer provided and launch exception if not.
  * Must be invoked from 'updateBasic() and decode()' at derived diameter types, when buffer should be printable
  *
  * @param buffer Raw avp data
  * @param size Raw avp data length
  *
  * @return Printable string or <null> if not printable
  */
  std::string assertPrintable(const char* buffer, const int size) const throw(anna::RuntimeException) {
    std::string result;

    if(size == 0) return result;

    bool printable;
    result = anna::functions::asAsciiString(buffer, size, printable);

    if(!printable) {
      std::string ex = getFormatName();
      ex += "::assertPrintable | Non-printable data provided";
      throw anna::RuntimeException(ex, ANNA_FILE_LOCATION);
    }

    return result;
  }

public:

  /**
  * Default constructor
  */
  AvpData() {};


  // gets

  /**
  * Gets the avp data format name
  *
  * @return avp data format name
  */
  virtual std::string getFormatName() const throw() = 0;

  /**
  * Gets the avp data size based on basic container.
  * The AVP Data field is zero or more octets.
  *
  * @return avp data size
  */
  virtual int getSize() const throw() = 0;


  // helpers

  /**
  * Encodes avp data part over buffer externally allocated
  *
  * @param buffer Raw data to be encoded
  * @param size Size of raw data to be encoded
  */
  void code(char* buffer, int &size) throw(anna::RuntimeException) {
    codeBasic(buffer, size);
  }


  /**
  * Gets the natural/smart string representation for avp data (format-dependent content)
  * Used in diameter message 'data' field
  * Default implementation launch exception when data is not printable
  *
  * @return Natural/smart string representation for avp data
  */
  virtual std::string asPrintableString() throw(anna::RuntimeException) {
    int size = getSize();
    char buffer[size];
    code(buffer, size);
    return (assertPrintable(buffer, size));
  }

  /**
  * Gets DataBlock binary block and ascii representation
  *
  * @return String with DataBlock representation
  */
  std::string asDataBlockString() throw(anna::RuntimeException) {
    int size = getSize();
    char buffer[size];
    code(buffer, size);
    anna::DataBlock db(buffer, size);
    return(db.asString());
  }

  /**
  * Class string representation
  * Default implementation invokes raw DataBlock 'asString' method,
  * but it should be different specially with complex application data types.
  *
  * @return String with class content
  */
  virtual std::string asString() throw(anna::RuntimeException) {
    return(asDataBlockString());
  }

  /**
  * Gets the hexadecimal string representation for avp data
  * Used in diameter message 'hex-data' field
  *
  * @return Hexadecimal string representation for avp data
  */
  std::string asHexString() throw(anna::RuntimeException) {
    int size = getSize();
    char buffer[size];
    code(buffer, size);
    anna::DataBlock db(buffer, size);
    return anna::functions::asHexString(db);
  }


  // sets

  /**
  * Decodes provided buffer/size
  *
  * Derived types must invoke base class 'decode()' at the end in order to keep coherence with parent members.
  * This base class decodification actually sets the base class members with the same buffer provided, being
  * more comfortable than using base class setters.
  *
  * @param buffer Raw avp data
  * @param size Raw avp data length
  */
  virtual void decode(const char* buffer, const int size) throw(anna::RuntimeException) = 0;

  /**
  * Initializes members from natural/smart string representation
  *
  * @param printableString avp data in natural/smart string representation (human-readable)
  */
  void fromPrintableString(const char * printableString) throw(anna::RuntimeException) {
    if(!printableString) {
      std::string ex = getFormatName();
      ex += "::fromPrintableString | Null printableString provided";
      throw anna::RuntimeException(ex, ANNA_FILE_LOCATION);
    }

    /*std::string dummy =*/assertPrintable(printableString, strlen(printableString));
    setPrintableString(printableString);
  }

  /**
  * Initializes members from hexadecimal string representation. I.e.: af1233fb01 (even number of digits).
  *
  * @param hexString Raw avp data in hexadecimal string representation
  */
  void fromHexString(const std::string& hexString) throw(anna::RuntimeException) {
    anna::DataBlock db(true);
    anna::functions::fromHexString(hexString, db);
    decode(db.getData(), db.getSize());
  }
};

}
}
}
}


#endif
