// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_stack_Engine_hpp
#define anna_diameter_stack_Engine_hpp


// Local
#include <anna/diameter/stack/Dictionary.hpp>

#include <anna/xml/DTDMemory.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/core/Singleton.hpp>

// STL
#include <string>
#include <vector>


namespace anna {

namespace diameter {

namespace stack {



//------------------------------------------------------------------------------
//----------------------------------------------------------------- class Engine
//------------------------------------------------------------------------------
/**
* Engine information
*/
class Engine : public anna::Singleton <Engine> {


  const anna::xml::DTDMemory * getDictionariesDTD() const throw() { return &a_dtd; }

public:

  typedef std::map<unsigned int, Dictionary*> stack_container;
  typedef stack_container::const_iterator const_stack_iterator;
  typedef stack_container::iterator stack_iterator;


  // get
  /**
  * Returns the dictionary registered with the provided identifier, which commonly
  * is the 'Application-Id'.
  *
  * @param stackId Stack identifier.
  * @return Dictionary reference, NULL if no stack found
  */
  const Dictionary * getDictionary(unsigned int stackId) const throw();

  /** Beginning stack iterator */
  const_stack_iterator stack_begin() const throw() { return a_stacks.begin(); }
  /** Ending stack iterator */
  const_stack_iterator stack_end() const throw() { return a_stacks.end(); }
  /** Stack size */
  int stack_size() const throw() { return a_stacks.size(); }

  // helpers
  /**
  * Gets boolean about empty stack condition
  *
  * @return Boolean about empty stack condition
  */
  bool isEmpty(void) const throw() { return (!a_stacks.size()); }

  /**
  * Class string representation
  *
  * @param all Complete engine information versus only stacks list with its ids and dictionary names
  *
  * @return String with class content
  */
  std::string asString(bool all = true) const throw();

  // set

  /**
  * Creates a diameter dictionary based on xml document file and store it over the provided stack identifier.
  * If the stack is currently used, an exception will be launched to notify the stack collision (a dictionary
  * could be removed with #removeStack).
  * If missing xml path file, the dictionary will be created without loading data.
  *
  * <pre>
  * Method 1 - Create and load xml data (commonly used for mono-load stacks):
  *     engine->createDictionary(MMS_STACK_ID, "/var/tmp/mms.xml");
  *
  * Method 2 - Create and then, load xml data (useful for multi-load stacks):
  *     Dictionary * d = engine->createDictionary(MMS_STACK_ID);
  *     // Loading data:
  *     d->load("/var/tmp/mms.xml");
  *     ...
  * Method 3 - Create and then, load data through Dictionary API (*):
  *     Dictionary * d = engine->createDictionary(MMS_STACK_ID);
  *     // Loading data:
  *     d->addFormat(...);
  *     d->addVendor(...);
  *     ...
  *
  * (*) Applications could create a dictionary with its own API (#Dictionary) but it is much more difficult,
  *     and usually, the method version with xml document load is used instead of this.
  *
  * </pre>
  *
  * @param stackId Stack identifier for created dictionary. We recommend to use the Diameter 'Application-Id' unless
  * the application is going to use dictionaries covering different applications (which is not very usual).
  * @param xmlPathFile Path file to the xml document which represents the diameter dictionary.
  *
  * @return Dictionary registered. When exception happen, dictionary can be accessed by #getDictionary
  */
  Dictionary * createDictionary(unsigned int stackId, const std::string & xmlPathFile = "") throw(anna::RuntimeException);

  /**
  * Register a externally created Dictionary or a derived class from Dictionary
  *
  * @param stackId Stack identifier for provided dictionary. We recommend to use the Diameter 'Application-Id' unless
  * the application is going to use dictionaries covering different applications (which is not very usual).
  * @param dictionary Externally created dictionary
  *
  * @return Dictionary registered. When exception happen, dictionary can be accessed by #getDictionary
  */
  Dictionary * registerDictionary(unsigned int stackId, Dictionary *dictionary) throw(anna::RuntimeException);

  /**
  * Loads an XML dictionary document over the diameter stack identifiers (one or more stack id's).
  * Passing more than one stack id, synchronized loadings are performed, which could be useful when
  * an application loads many dictionaries with common avps.
  *
  * @param stacks Stacks identifiers over which the dictionary will be load.
  * @param xmlPathFile Path file to the xml document which represents the diameter dictionary.
  */
  void loadDictionary(const std::vector<unsigned int> & stacks, const std::string & xmlPathFile) throw(anna::RuntimeException);

  /**
  * Loads an XML dictionary document over all the diameter engine registered stacks. When more than one stack id is
  * used, synchronized loadings are performed, which could be useful when an application loads many dictionaries with
  * common avps.
  *
  * @param xmlPathFile Path file to the xml document which represents the diameter dictionary.
  */
  void loadDictionary(const std::string & xmlPathFile) throw(anna::RuntimeException);

  /**
  * Remove all stacks registered on diameter stack engine
  */
  void removeStacks(void) throw() { a_stacks.clear(); }

  /**
  * Remove the stack provided.
  *
  * @param stackId Stack identifier for created dictionary
  */
  void removeStack(unsigned int stackId) throw();

private:

  anna::xml::DTDMemory a_dtd;
  stack_container a_stacks;

  Engine();

  friend class anna::Singleton <Engine>;
  friend class Dictionary; // access to 'getDictionariesDTD()'
};


}
}
}


#endif
