// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_timex_TimeEvent_hpp
#define anna_timex_TimeEvent_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/config/defines.hpp>
#include <anna/core/util/Millisecond.hpp>

namespace anna {

namespace timex {

class Engine;
class TimeEventObserver;

/**
   Clase base de los eventos de los eventos de tiempo.
*/
class TimeEvent {
public:
  /**
     Sinonimo usado para el identificador de eventos de tiempo.

     Se define como un ptrnumber para facilitar la construccion de instancias sin tener que indicar
     un Id en particular, ya que se usara su direccion de memoria como identificador.
  */
  typedef anna::ptrnumber Id;

  /**
     Destructor.
     Si un evento de tiempo esta activo y se invoca a su destructor se invocaria automaticamente a
     Engine::cancel para terminar con la ejecucin de este evento.
  */
  virtual ~TimeEvent();

  /**
     Devuelve el identificador de este evento.
     \return El identificador de este evento.
  */
  Id getId() const throw() { return a_id; }

  /**
     Devuelve la duracion maxima a este evento indicada en milisegundos.
     @return La duracion maxima de este evento indicada en milisegundos.
  */
  const Millisecond & getTimeout() const throw() { return a_timeout; }

  /**
     Devuelve la instancia del objeto encargado de gestionar el espacio asociado a
     esta instancia. Puede ser NULL.
     \return La instancia del objeto encargado de gestionar el espacio asociado a
     esta instancia.
  */
  const TimeEventObserver* getObserver() const throw() { return a_observer; }

  /**
     Establece el identificador de este evento de tiempo.
     \param id El identificador de evento para esta instancia.

     \warning Exclusivamente uso interno.
  */
  void setId(const Id id) throw() { a_id = id; }

  /**
     Establece la duracion de este evento indicada en milisegundos.
     \param timeout Establece la duracion de este evento.
  */
  void setTimeout(const Millisecond & timeout) throw() { a_timeout = timeout; }

  /**
     Establece la instancia de objeto encargado de gestionar el espacio asociado a esta
     instancia. Si no es NULL sera invocado cuando el timex::Engine caduque o
     cancele este evento temporal.

     \param observer Instancia del objeto encargada de gestional el espacio asignado a esta
     instancia.
  */
  void setObserver(TimeEventObserver* observer) throw() { a_observer = observer; }

  /**
     Devuelve el estado de activacin de este evento de tiempo.
     @return \em false si el evento esta activado o \em false en otro caso.
  */
  bool isActive() const throw() { return (a_controller != NULL); }

  /**
     Devuelve una cadena con la informacion sobre este evento.
     \return Una cadena con la informacion sobre este evento.
  */
  virtual std::string asString() const throw();

protected:
  /**
     Contructor.
  */
  TimeEvent() :
    a_id(0),
    a_timeout(0),
    a_controller(NULL),
    a_observer(NULL)
  {;}

  /**
     Contructor.

     @param id Identificador de este evento.
     @param timeout La duracion de este evento indicada en milisegundos.
  */
  TimeEvent(const Id id, const Millisecond & timeout) :
    a_id(id),
    a_timeout(timeout),
    a_controller(NULL),
    a_observer(NULL) {}

  /**
     metodo que debemos re-escribir para particularizar el comportamiento de nuestra clase cuando
     el Engine notifica que se ha sobrepasado la duracion maxima del evento si que se invoque
     a ninguno de los metodos que lo cancelarian. Una vez invocado a este metodo el evento se considera
     cancelado y todos sus recursos son liberados automaticamente por Engine.

     @param timeController Controlador de tiempo asociado al evento que esta expirando.
  */
  virtual void expire(Engine* timeController) throw(RuntimeException) = 0;

  /**
     metodo que debemos re-escribir para particularizar el comportamiento de nuestra clase cuando
     el Engine notifica que se ha parado el sistema de control de tiempos. Una vez invocado a este
     metodo el evento se considera cancelado y todos sus recursos son liberados automaticamente
     por Engine.

     Las acciones realizadas nunca deber�n generar nuevos eventos. Por defecto no hace nada.
  */
  virtual void stop() throw(RuntimeException) {;}

private:
  Id a_id;
  Millisecond a_timeout;
  Engine* a_controller;
  TimeEventObserver* a_observer;

  friend class Engine;
};

}
}

#endif


