// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <algorithm>

#include <anna/core/functions.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/core/tracing/TraceFunction.hpp>
#include <anna/core/mt/Thread.hpp>

#include <anna/app/functions.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

#include <anna/comm/Communicator.hpp>

#include <anna/timex/Engine.hpp>
#include <anna/timex/internal/TickConsumer.hpp>
#include <anna/timex/internal/TickProducer.hpp>
#include <anna/timex/TimeEventObserver.hpp>
#include <anna/timex/internal/sccs.hpp>

using namespace std;
using namespace anna;
using namespace anna::comm;

/**
   Minimum resolution (in milliseconds) supported by the time manager
*/
//static
const Millisecond anna::timex::Engine::minResolution(10);


anna::timex::Engine::Engine(const Millisecond & maxTimeout, const Millisecond & resolution) :
  app::Component(getClassName()),
  a_currentQuantum(0),
  a_maxQuantum(0),
  a_maxTimeout(maxTimeout),
  a_resolution(resolution),
  a_isActive(false),
  a_tickProducer(NULL),
  a_tickConsumer(NULL),
  a_expiredQuantum(NULL),
  a_timeTable(NULL) {
  timex::sccs::activate();
}

anna::timex::Engine::~Engine() {
  delete a_tickConsumer;
  delete [] a_timeTable;
}

//--------------------------------------------------------------------------------------------
// Inicializa la configuracin.
//
// (2) En Solaris el ualarm tiene una deriva constante que parece ser que dependen de la
//     arquitectura de la m�uina. Para calcular esta deriva y poder corregirla en los
//     posteriores ualarm vamos a calcular la diferencia entre el tiempo esperado y el momento
//     en que realmente llega la seal de ualarm.
//--------------------------------------------------------------------------------------------
void anna::timex::Engine::do_initialize()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tm("timex::Engine", "do_initialize", ANNA_FILE_LOCATION));

  if(a_maxQuantum > 0) {
    Logger::write(Logger::Warning, "Time controller was previously started", ANNA_FILE_LOCATION);
    return;
  }

  if(a_resolution < minResolution)
    throw RuntimeException(functions::asString("Resolution must be greater than %d milliseconds", minResolution.getValue()), ANNA_FILE_LOCATION);

  if(a_resolution < (Millisecond)100) {
    LOGWARNING(Logger::warning(functions::asString("Resolutions under 100 milliseconds (%d in this case) slightly increase the CPU usage", a_resolution.getValue()), ANNA_FILE_LOCATION));
  }

  if(a_maxTimeout <= a_resolution)
    throw RuntimeException(functions::asString("Max-Timeout must be greater than %d milliseconds", a_resolution.getValue()), ANNA_FILE_LOCATION);

  a_maxQuantum = a_maxTimeout / a_resolution;

  while((a_maxQuantum * a_resolution) <= a_maxTimeout) a_maxQuantum ++;

  a_timeTable = new Quantum [a_maxQuantum];
  Communicator* communicator = app::functions::component <Communicator> (ANNA_FILE_LOCATION);
  communicator->attach(a_tickConsumer = new TickConsumer(this));
  /**
   * Esto siempre se ejecutar en un thread aparte aunque la librera haya sido generada en modo ST,
   * porque as evitamos tener que generar el pulso de reloj mediante una alarma.
   */
  a_tickProducer = new TickProducer(this, a_tickConsumer->getfdWrite());
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  int errorCode;

  if((errorCode = pthread_create(&a_threadProducer, &attr, TickProducer::exec, a_tickProducer)) != 0)
    throw RuntimeException(std::string("timex::Engine::do_initialize"), errorCode, ANNA_FILE_LOCATION);

  LOGDEBUG(
    string msg("Time controller | Max Timeout: ");
    msg += functions::asString(a_maxTimeout);
    msg += " | Resolution: ";
    msg += functions::asString(a_resolution);
    msg += " | Max Quantum: ";
    msg += functions::asString(a_maxQuantum);
    msg += " quantums";
    Logger::write(Logger::Debug, msg, ANNA_FILE_LOCATION);
  );
}

// Reimplementado de app::Component
void anna::timex::Engine::do_cloneParent()
throw() {
}

/*
 * Se invoca desde app::Application::clone -> app::Component::do_cloneChild (ojo EN EL NUEVO PROCESO).
 * Instala la senhal de tick en el proceso, ya que la alarma no se hereda directamente.
 */
void anna::timex::Engine::do_cloneChild()
throw(RuntimeException) {
}

//----------------------------------------------------------------------------
// No para los hilos de generacion, sino que evita que se escriban los bytes
// en el 'pipe'.
//----------------------------------------------------------------------------
void anna::timex::Engine::pause()
throw(RuntimeException) {
  Guard guard(this, "timex::Engine (pause)");

  if(a_tickProducer->isInPause() == false) {
    LOGWARNING(
      string msg("timex::Engine::pause | Pending TimeEvents: ");
      msg += functions::asString((const int) a_directory.size());
      Logger::warning(msg, ANNA_FILE_LOCATION);
    );
    a_tickProducer->setIsInPause(true);
  }
}

void anna::timex::Engine::play()
throw(RuntimeException) {
  Guard guard(this, "timex::Engine (play)");

  if(a_tickProducer->isInPause() == true) {
    LOGWARNING(
      string msg("timex::Engine::play | Pending TimeEvents: ");
      msg += functions::asString((const int) a_directory.size());
      Logger::warning(msg, ANNA_FILE_LOCATION);
    );
    a_tickProducer->setIsInPause(false);
  }
}

void anna::timex::Engine::activate(timex::TimeEvent* timeEvent)
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tm(Logger::Local7, "timex::Engine", "activate", ANNA_FILE_LOCATION));

  if(a_maxQuantum == 0)
    throw RuntimeException("Engine::initialize was not called", ANNA_FILE_LOCATION);

  if(timeEvent == NULL)
    throw RuntimeException("Cannot activate a NULL TimeEvent", ANNA_FILE_LOCATION);

  if(timeEvent->a_controller != NULL) {
    string msg(timeEvent->asString());
    msg += " | Already activated";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  const Millisecond & timeout(timeEvent->getTimeout());

  if(timeout > a_maxTimeout || timeout < a_resolution) {
    string msg("Invalid TimeEvent timeout | Max Timeout: ");
    msg += functions::asString(a_maxTimeout);
    msg += " | Min timeout (resolution): ";
    msg += functions::asString(a_resolution);
    msg += " | ";
    msg += timeEvent->asString();
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  Guard guard(this, "timex::Engine (activate)");

  if(a_tickProducer->isInPause() == true) {
    string msg("Cannot activate TimeEvents with timex::Engine in pause | ");
    msg += timeEvent->asString();
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  const TimeEvent::Id id(timeEvent->getId());

  if(a_directory.find(id) != a_directory.end())
    throw RuntimeException(functions::asString("Id %ld (0x%x) already in use", id, id), ANNA_FILE_LOCATION);

  int iq = getQuantum(timeout);
  Quantum* quantum = &a_timeTable [iq];
  bool delayed;
  timeEvent->a_controller = this;
  a_directory.insert(Directory::value_type(id, quantum));

  if(quantum == a_expiredQuantum) {
    a_delayedQuantum.push_back(timeEvent);
    delayed = true;
  } else {
    quantum->push_back(timeEvent);
    delayed = false;
  }

  LOGDEBUG(
    string msg("timex::Engine::activate | ");
    msg += timeEvent->asString();
    msg += " | Current quantum: ";
    msg += functions::asString(a_currentQuantum);
    msg += " | Quantum programmed: ";
    msg += functions::asString(iq);
    msg += " (";
    msg += functions::asHexString(anna_ptrnumber_cast(quantum));
    msg += functions::asText(") | Delayed: ", delayed);
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

anna::timex::TimeEvent* anna::timex::Engine::getTimeEvent(const timex::TimeEvent::Id eventTimeId)
throw() {
  LOGMETHOD(TraceMethod tm(Logger::Local7, "timex::Engine", "getTimeEvent", ANNA_FILE_LOCATION));
  Directory::iterator iid;
  TimeEvent* result(NULL);
  Quantum* quantum;
  TimeEvent* aux;
  Guard guard(this, "timex::Engine (getTimeEvent)");

  if((iid = a_directory.find(eventTimeId)) == a_directory.end())
    return NULL;

  quantum = iid->second;

  for(Quantum::iterator ii = quantum->begin(), maxii = quantum->end(); ii != maxii; ii ++) {
    if((aux = *ii)->getId() == eventTimeId) {
      result = aux;
      break;
    }
  }

  return result;
}

void anna::timex::Engine::cancel(timex::TimeEvent* timeEvent)
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tm(Logger::Local7, "timex::Engine", "cancel", ANNA_FILE_LOCATION));

  if(timeEvent == NULL)
    throw RuntimeException("Cannot cancel a NULL TimeEvent", ANNA_FILE_LOCATION);

  if(timeEvent->a_controller == NULL) {
    LOGDEBUG(
      string msg(timeEvent->asString());
      msg += " | Not activated";
      Logger::write(Logger::Debug, msg, ANNA_FILE_LOCATION);
    );
    return;
  }

  Directory::iterator iid;
  Quantum* quantum;
  Guard guard(this, "timex::Engine (cancel)");

  if((iid = a_directory.find(timeEvent->getId())) == a_directory.end())
    return;

  if((quantum = iid->second) == a_expiredQuantum) {
    LOGWARNING(
      string msg("timex::Engine::cancel | ");
      msg += timeEvent->asString();
      msg += " | Processing already programmed";
      Logger::warning(msg, ANNA_FILE_LOCATION);
    );
    return;
  }

  a_directory.erase(iid);
  quantum->erase(find(quantum->begin(), quantum->end(), timeEvent));
  timeEvent->a_controller = NULL;
  LOGDEBUG(
    string msg("timex::Engine::cancel | ");
    msg += timeEvent->asString();
    msg += " | Current quantum: ";
    msg += functions::asString(a_currentQuantum);
    msg += " | Quantum programmed: ";
    msg += functions::asHexString(anna_ptrnumber_cast(quantum));
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  notifyRelease(timeEvent);
}

void anna::timex::Engine::do_stop()
throw() {
  LOGMETHOD(TraceMethod tm("timex::Engine", "do_stop", ANNA_FILE_LOCATION));
  Quantum::iterator ii, maxii;
  Guard guard(this, "timex::Engine (do_stop)");
  app::functions::component <Communicator> (ANNA_FILE_LOCATION)->detach(a_tickConsumer);

  if(a_tickProducer)
    a_tickProducer->requestStop();

  a_directory.clear();
  TimeEvent* timeEvent;

  for(int iq = 0; iq < a_maxQuantum; iq ++) {
    Quantum& quantum(a_timeTable [iq]);

    for(ii = quantum.begin(), maxii = quantum.end(); ii != maxii; ii ++) {
      timeEvent = Engine::timeEvent(ii);

      try {
        timeEvent->a_controller = NULL;
        timeEvent->stop();
        notifyRelease(timeEvent);
      } catch(Exception& ex) {
        ex.trace();
      }
    }

    quantum.clear();
  }
}

void anna::timex::Engine::kill()
throw() {
  Guard guard(this, "timex::Engine (kill)");
  app::functions::component <Communicator> (ANNA_FILE_LOCATION)->detach(a_tickConsumer);

  if(a_tickProducer)
    a_tickProducer->requestStop();

  string msg("timex::Engine::kill | ");
  msg += functions::asString(" | Pending TimeEvents: %d", a_directory.size());
  Logger::critical(msg, ANNA_FILE_LOCATION);

  for(int iq = 0; iq < a_maxQuantum; iq ++)
    a_timeTable [iq].clear();

  a_delayedQuantum.clear();
  a_directory.clear();
}

//----------------------------------------------------------------------------------
// (1) Podr� ser que al invocar al m�odo 'expire' de una evento se halla originado
//     la cancelacin de algn otro evento. As�que si el id de evento ya no esta en
//     en el directorio => da igual, se ignora ya que luego borraremos todos los
//     eventos del Quantum.
// (2) Si mientas estamos caducando este Quantum hay que activar transacciones que
//     debe ir sobre el => se guardan en un vector temporal y se copian al final
//     del proceso, con lo que evita totalmente la posiblidad de perdida.
// (3) Si el temporizador ha sido reactivado no tiene que liberarlo.
//----------------------------------------------------------------------------------
void anna::timex::Engine::tick()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tm(Logger::Local7, "timex::Engine", "tick", ANNA_FILE_LOCATION));
  /*
     if (Logger::isActive (Logger::Local6) == true && a_directory.size () > 0) {
        string msg ("Tick | Quantum: ");
        msg += functions::asString (a_currentQuantum);
        msg += " | Tx: ";
        msg += functions::asString (functions::millisecond ());
        msg += " | N: ";
        msg += functions::asString (a_directory.size ());
        Logger::write (Logger::Local6, msg, ANNA_FILE_LOCATION);
     };
  */
  Guard guard(this, "timex::Engine (tick)");
  int counter(0);
  eventBeginQuantum();
  a_expiredQuantum = &a_timeTable [a_currentQuantum];
  TimeEvent* timeEvent;
  Directory::iterator iid;
  quantum_iterator ii, maxii;
  bool observed;

  for(ii = a_expiredQuantum->begin(), maxii = a_expiredQuantum->end(); ii != maxii; ii ++) {
    timeEvent = Engine::timeEvent(ii);

    if((iid = a_directory.find(timeEvent->getId())) == a_directory.end())            // (1)
      continue;

    a_directory.erase(iid);

    if(Logger::isActive(Logger::Debug)) {
      string msg(timeEvent->asString());
      msg += " | Quantum programmed: ";
      msg += functions::asHexString(anna_ptrnumber_cast(a_expiredQuantum));
      msg += " | Expired";
      Logger::write(Logger::Debug, msg, ANNA_FILE_LOCATION);
    }

    timeEvent->a_controller = NULL;
    /*
     * Si tiene NO tiene observer asociado, seguramente se auto-liberar en el expire
     * Si tiene observer, se invocar al mtodo correspondiente.
     */
    observed = (timeEvent->a_observer != NULL);

    try {
      timeEvent->expire(this);
      counter ++;
    } catch(Exception& ex) {
      ex.trace();
    }

    // Si tiene observer => el expire NO puede haber liberado el TimeEvent
    if(observed == true) {
      if(timeEvent->a_controller == NULL)                                           // (3)
        timeEvent->a_observer->release(timeEvent);
    }
  }

  a_expiredQuantum->clear();

  if(a_delayedQuantum.size() > 0) {                                                 // (2)
    for(ii = a_delayedQuantum.begin(), maxii = a_delayedQuantum.end(); ii != maxii; ii ++)
      a_expiredQuantum->push_back(Engine::timeEvent(ii));

    a_delayedQuantum.clear();
  }

  a_expiredQuantum = NULL;

  if(++ a_currentQuantum >= a_maxQuantum)
    a_currentQuantum = 0;

  eventEndQuantum(counter);
}

void anna::timex::Engine::notifyRelease(timex::TimeEvent* timeEvent)
throw() {
  if(timeEvent->a_observer != NULL)
    timeEvent->a_observer->release(timeEvent);
}

string anna::timex::Engine::asString() const
throw() {
  string msg("timex::Engine { ");
  msg += app::Component::asString();
  msg += " | Max Timeout: ";
  msg += functions::asString(a_maxTimeout);
  msg += " ms | Resolution: ";
  msg += functions::asString(a_resolution);
  msg += " ms | Programming: ";
  msg += functions::asString(a_maxQuantum);
  msg += " quantums | Pending TimeEvents: ";
  msg += functions::asString((const int) a_directory.size());
  return msg += " }";
}

xml::Node* anna::timex::Engine::asXML(xml::Node* parent) const
throw() {
  parent = Component::asXML(parent);
  xml::Node* result = parent->createChild("timex.Engine");
  result->createAttribute("MaxTimeout", a_maxTimeout);
  result->createAttribute("Resolution", a_resolution);
  result->createAttribute("N", a_directory.size());
  result->createAttribute("Pause", functions::asString(a_tickProducer->isInPause()));
  xml::Node* node = result->createChild("Quantum");
  node->createAttribute("Current", a_currentQuantum);
  node->createAttribute("Max", a_maxQuantum);
  return result;
}
